
SetLogFile("examples/globalrep-example.log");
GetVersion();
Attach("globalrep.m");
// compute global representations for all S_3-extensions of Q_3
GlobalRepresentations( SymmetricGroup(3), 3 );
// compute global representations for all S_3- and C_3-extensions of Q_3
// and include corresponding polynomial from the database by Jones/Roberts
GlobalRepresentations( [< SymmetricGroup(3), 3 >, <CyclicGroup(6), 3> ] : JR);

// for extensions of degree 8 of Q_2 (or degree 9 of Q_3), the
// above method is slow.
// Instead, use the list of polynomials from the database by Jones/Roberts
// (which contains informations about Galois groups) and compute global
// representations for each of those polynomials.

SetVerbose("GlobalRepresentations",1);

// initialize polynomial variables used in the database
Zy<y> := PolynomialRing(Integers());
Zt<t> := PolynomialRing(Integers());
Zx<x> := PolynomialRing(Integers());
// load polynomials into the variable pols
load "JR/Q2deg8a.m";
// compute representations (takes about 30 seconds)
GlobalRepresentationsJR(pols, 2);

// the same for degree 9 extensions of Q_3 (takes a few seconds)
load "JR/Q3deg9a.m";
time GlobalRepresentationsJR(pols, 3);


SetVerbose("GlobalRepresentations", 0);

// all groups up to degree 15 with primes dividing the group order:
// (these are ordered in such a way that for every degree and primes
// the command AllExtensions is just called once)
groups := [  <SmallGroup(n,i), p> : i in [1..NumberOfSmallGroups(n)],
                                    p in [x[1] : x in Factorization(n)],
                                    n in [2..15] ];
// representations for extensions of degree 8 and 9 have already been
// computed by the commands above
degree12 := [x : x in groups | #x[1] eq 12];
degree14 := [x : x in groups | #x[1] eq 14];
rest := [x : x in groups | not #x[1] in {8,9,12,14} ];

// Representations up to degree 15, excluding n=8,9,12,14
// takes about 3 minutes
time repRest := GlobalRepresentations(rest : JR);

// Representation for degree 14 extensions
// takes about 3 minutes
time rep14 := GlobalRepresentations(degree14);

// Representation for degree 12 extensions
// takes about 5 hours
time rep12 := GlobalRepresentations(degree12);



