
/***************************************************
 * Computation of the local fundamental class
 * using the direct method (Example 2.6)
 ***************************************************/

SetLogFile("lfc-direct.log");
GetVersion();

Attach("../magma/brauer.m");
Zx<x> := PolynomialRing(Integers());


// S3
Qp := pAdicField(3,30);
L := ext<Qp | x^6+3>;
time LocalFundamentalClassDirect(L, 10);
// --> Time: 24.420
Qp := pAdicField(3,50);
L := ext<Qp | x^6+3>;
time LocalFundamentalClassDirect(L, 10);
// --> Time: 26.530
Qp := pAdicField(3,80);
L := ext<Qp | x^6+3>;
time LocalFundamentalClassDirect(L, 10);
// --> Time: 26.620
Qp := pAdicField(3,100);
L := ext<Qp | x^6+3>;
time LocalFundamentalClassDirect(L, 10);
// --> Time: 27.220

Qp := pAdicField(3,30);
L := ext<Qp | x^6+3>;
time LocalFundamentalClassDirect(L, 20);
// --> Time: 85.790
Qp := pAdicField(3,50);
L := ext<Qp | x^6+3>;
time LocalFundamentalClassDirect(L, 20);
// --> Time: 88.610




// D4
L := NumberField(x^8 + 38*x^4 + 1);
P := Factorization(2*RingOfIntegers(L))[1,1];
K := Completion(L,P);
L := ChangePrecision(K, 240);
time LocalFundamentalClassDirect(L, 10);
// --> Time: 435.010 = 7.2 min
L := ChangePrecision(K, 400);
time LocalFundamentalClassDirect(L, 10);
// --> Time: 724.620 = 12 min
L := ChangePrecision(K, 640);
time LocalFundamentalClassDirect(L, 10);
// --> Time: 746.390 = 12.4 min
L := ChangePrecision(K, 800);
time LocalFundamentalClassDirect(L, 10);
// --> Time: 787.320 = 13 min

L := ChangePrecision(K, 240);
time LocalFundamentalClassDirect(L, 20);
// --> Time: 1831.000 = 30.5 min
L := ChangePrecision(K, 400);
time LocalFundamentalClassDirect(L, 20);
// --> Time: 4481.940 = 74.7 min / 1.2 h


// D5
L := NumberField(x^10 - 10*x^8 + 30*x^7 + 90*x^6 - 162*x^5 + 125*x^4 + 90*x^3 - 80*x^2 - 120*x + 144);
P := Factorization(5*RingOfIntegers(L))[1,1];
K := Completion(L,P);
L := ChangePrecision(K, 150);
time LocalFundamentalClassDirect(L, 10);
// --> Time: 6176.190  = 102.9 min / 1.7 h 
L := ChangePrecision(K, 250);
time LocalFundamentalClassDirect(L, 10);
// --> Time: 10291.100 = 171.5 min / 2.9 h
L := ChangePrecision(K, 400);
time LocalFundamentalClassDirect(L, 10);
// --> Time: 17289.690 = 288.2 min / 4.8 h
L := ChangePrecision(K, 500);
time LocalFundamentalClassDirect(L, 10);
// --> Time: 22438.210 = 374 min / 6.2 h

L := ChangePrecision(K, 250);
time LocalFundamentalClassDirect(L, 20);
// --> Time: 28961.090 = 482.7 min / 8 h




/***************************************************
 * ATTENTION !!
 * For this example in degree 12, Magma sometimes refuses
 * to compute the automorphism group. One then needs to
 * restart Magma and try again.
 ***************************************************/
Attach("../magma/brauer.m");
Zx<x> := PolynomialRing(Integers());

// Q12
f := x^12 - 6*x^11 - 30*x^10 + 190*x^9 + 171*x^8 - 1740*x^7 + 124*x^6 + 6420*x^5 - 2409*x^4 - 9630*x^3 + 3330*x^2 + 5214*x - 659;
L := NumberField(f);
P := Factorization(3*RingOfIntegers(L))[1,1];
K := Completion(L,P);
L := ChangePrecision(K, 90);
time LocalFundamentalClassDirect(L, 10);
// --> Time: 2031.090 = 33.8 min
L := ChangePrecision(K, 150);
time LocalFundamentalClassDirect(L, 10);
// --> Time: 3379.750 = 56.3 min
L := ChangePrecision(K, 240);
time LocalFundamentalClassDirect(L, 10);
// --> Time: 5708.740 = 95.1 min / 1.6 h
L := ChangePrecision(K, 300);
time LocalFundamentalClassDirect(L, 10);
// --> Time: 7192.580 = 119.9 min / 2 h

L := ChangePrecision(K, 150);
time LocalFundamentalClassDirect(L, 20);
// --> Time: 9707.380 = 161.8 min / 2.7 h



