
/***************************************************
 * Solving Frobenius Equations
 * (Example 2.11)
 ***************************************************/

SetLogFile("lfc-frobeq.log");
GetVersion();

Attach("../magma/brauer.m");

Zx<x> := PolynomialRing(Integers());
Q3 := pAdicField(3,100);

L := ext<Q3 | x^6+6*x^2+6>;
G, psi := AutomorphismGroup(L,Q3);
OL := RingOfIntegers(L);
pi := L!OL.1;

sig := [ g : g in G| Order(g) eq 3][1];
bool, c := IsExactlyDivisible( OL!psi(sig)(pi), OL!pi);

// Find solution of precision 5
u, frob := FrobeniusEquation(c, 5);
Degree(Parent(u), OL);

// Find solution of precision 20
time u, frob := FrobeniusEquation(c, 20);
Degree(Parent(u), OL);


// For elements of norm 1, solutions can be found
// a lot faster, and without producing extensions
OL1 := UnramifiedExtension(OL, 3);
c1 := -OL1.1;
Norm(c1, OL);
time FrobeniusEquation(c1, 500, OL);


