
/***************************************************
 * Computation of the local fundamental class
 * using Serre's approach (Example 2.18)
 ***************************************************/

SetLogFile("lfc-serre.log");
GetVersion();

AttachSpec("../magma/brauer.m");
Zx<x> := PolynomialRing(Integers());

// S3
Qp := pAdicField(3,50);
L := ext<Qp | x^6+3>;
time LocalFundamentalClassSerre(L, Qp, 10);
// --> Time: 1.250 = 0.02 min
time LocalFundamentalClassSerre(L, Qp, 20);
// --> Time: 1.270 = 0.02 min


// D4
L := NumberField(x^8 + 38*x^4 + 1);
P := Factorization(2*RingOfIntegers(L))[1,1];
K := Completion(L,P);
L := ChangePrecision(K, 400);
time LocalFundamentalClassSerre(L, pAdicField(L), 10);
// --> Time: 96.200 = 1.6 min
time LocalFundamentalClassSerre(L, pAdicField(L), 20);
// --> Time: 96.250 = 1.6 min


// D5
L := NumberField(x^10 - 10*x^8 + 30*x^7 + 90*x^6 - 162*x^5 + 125*x^4 + 90*x^3 - 80*x^2 - 120*x + 144);
P := Factorization(5*RingOfIntegers(L))[1,1];
K := Completion(L,P);
L := ChangePrecision(K, 250);
time LocalFundamentalClassSerre(L, pAdicField(L), 10);
// --> Time: 880.660 = 14.7 min
time LocalFundamentalClassSerre(L, pAdicField(L), 20);
// --> Time: 885.530 = 14.8 min


// Q12
f := x^12 - 6*x^11 - 30*x^10 + 190*x^9 + 171*x^8 - 1740*x^7 + 124*x^6 + 6420*x^5 - 2409*x^4 - 9630*x^3 + 3330*x^2 + 5214*x - 659;
L := NumberField(f);
P := Factorization(3*RingOfIntegers(L))[1,1];
K := Completion(L,P);
L := ChangePrecision(K, 150);
time LocalFundamentalClassSerre(L, pAdicField(L), 10);
// --> Time: 1125.060 = 18.8 min
time LocalFundamentalClassSerre(L, pAdicField(L), 20);
// --> Time: 1127.070 = 18.8 min



