
SetLogFile("D7.log");
GetVersion();

AttachSpec("../magma/spec");

Zx<x> := PolynomialRing(Rationals());
G := DihedralGroup(7);


/***************************************************
 *     Verify conjecture for G=D7, p=2
 ***************************************************/

time pols := GlobalRepresentations( G, 2 );
pols;
// --> no extensions

/***************************************************
 *     Verify conjecture for G=D7, p=7
 ***************************************************/

// the following polynomials can be found by
// class field theory techniques
pols := [
    x^14 - 7*x^13 + 21*x^12 - 35*x^11 + 35*x^10 - 35*x^8 + 65*x^7 - 35*x^6 + 35*x^4 - 35*x^3 + 21*x^2 - 7*x + 1,
    x^14 + 14*x^12 + 63*x^10 + 728*x^8 + 7231*x^6 + 9702*x^4 + 3969*x^2 + 61236,
    x^14 - 112*x^12 + 2982*x^10 - 20608*x^8 + 55321*x^6 - 62496*x^4 + 27216*x^2 - 3456
];


L := NumberField(pols[1]);
lec := LECcreateRec(L, 7);
LECverify(~lec);

L := NumberField(pols[2]);
lec := LECcreateRec(L, 7);
// An error occurs when computing the units of the local field:
// Runtime error: PrincipalUnitGroup: ERROR in discrete logarithm (quadratic).
// To solve this, we initialize the completion with a higher precision
LECinitializeCompletion(~setting, 50);
LECverify(~lec);


t := Cputime();
L := NumberField(pols[3]);
setting := LECcreateRec(L, 7);
// In this example the global field E would have degree 588.
// To solve this one can compute the epsilon constants first (they will be rational)
// and use a smaller cyclotomic field for the other computations
time LECepsilonConstant(~setting);
setting`Qmpt := CyclotomicField(Exponent(setting`G));
setting`tLK := [* setting`Qmpt!Rationals()!x : x in setting`tLK *];
// now one can verify the conjecture
LECverify(~setting);
Cputime(t);

